

// Written by Hans de Nivelle,
// produces axioms that an addition game of the form
//    SEND
//    MORE
//   MONEY


#include <iostream>
#include <string>
#include <sstream>
#include <list>
#include <vector>
#include <map>


bool checkcapitals( const std::string& s )
{
   for( std::string::const_iterator
           p = s. begin( );
	   p != s. end( );
	   ++ p )
   {
      if( *p < 'A' || *p > 'Z' )
	 return false;
   }
   return true;
}


// Write two summation clauses expressing:
// If xi has value x, y1 has value y, ci has value c,
// then c[i+1] has value carry and z[i] has value sum.
// All of x,y,c can be < 0, in which case they are 
// undefined. 


void writesumclauses( unsigned int i, int x, int y, int c )
{
   unsigned int sum = 0;
   unsigned int carry = 0;

   if( c > 1 ) exit(0);
   if( i == 0 && c >= 0 ) exit(0);
      // At position 0, there can be no ingoing carry.

   if( x >= 0 )
      sum += x;
   if( y >= 0 )
      sum += y;
   if( c >= 0 )
      sum += c;

   if( sum >= 10 )
   {
      sum -= 10;
      ++ carry;
   }

   // We write two clauses, one for the sum and one for the carry.

   for( unsigned int kk = 0; kk < 2; ++ kk )
   {
      std::cout << "[ ";

      if( x >= 0 )
         std::cout << "-X" << i << x << ", ";
      if( y >= 0 )
         std::cout << "-Y" << i << y << ", ";
      if( c >= 0 )
         std::cout << "-C" << (i-1) << c << ", ";

      if( kk == 0 )
         std::cout << "Z" << i << sum;
      else
         std::cout << "C" << i << carry;

      std::cout << " ]\n";
   }
}


// Write a clause that s has at least one value.

void writeatleastone( const std::string& s, unsigned int r0, unsigned int r1 )
{

   std::cout << "[ ";
   for( unsigned int r = r0; r < r1; ++ r )
   {
      if( r != r0 )
         std::cout << ", ";

      std::cout << s << r;
   }

   std::cout << " ]\n";
}


void writeatmostone( const std::string& s, unsigned int r0, unsigned int r1 )
{
   for( unsigned int i = r0; i < r1; ++ i )
      for( unsigned int j = i + 1; j < r1; ++ j )
      {
         std::cout << "[ ";
	 std::cout << "-" << s << i << ", ";
	 std::cout << "-" << s << j << " ]\n";
      }
}


void writeequalityclauses( const std::string& s1, const std::string& s2 )
{
   for( unsigned int i = 0; i < 10; ++ i )
   {
      std::cout << "[ ";
      std::cout << "-" << s1 << i << ", ";
      std::cout << s2 << i;
      std::cout << " ]\n";
   }
}


void writedifferenceclauses( const std::string& s1, const std::string& s2 )
{
   for( unsigned int i = 0; i < 10; ++ i )
   {
      std::cout << "[ ";
      std::cout << "-" << s1 << i << ", ";
      std::cout << "-" << s2 << i << " ]\n";
   }
}


int main( unsigned int nrargs, char* ( args [] ))
{
   std::string s1;
   std::string s2;
   std::string s3;
      // We assume that the puzzle is s1 + s2 = s3.

   std::cerr << "first number:    ";
   std::cin >> s1;
   
   std::cerr << "second number:   ";
   std::cin >> s2;

   std::cerr << "sum:             ";
   std::cin  >> s3;


   std::cerr << s1 << "\n";
   std::cerr << s2 << " +\n";
   std::cerr << "============\n";
   std::cerr << s3 << "\n\n";


   // We make a few sanity checks.

   unsigned int l1 = s1. size( );
   unsigned int l2 = s2. size( );
   unsigned int l3 = s3. size( );

   if( l1 == 0 || l2 == 0 )

   {
      printf( "addend must be non empty!\n" );
      return 0;
   }

   unsigned int max_l1l2 = ( l1 > l2 ) ? l1 : l2;
   unsigned int min_l1l2 = ( l1 < l2 ) ? l1 : l2;

   if( l3 < max_l1l2 )
   {
      std::cerr << "result too short\n";
      return 0;
   }

   if( l3 > max_l1l2 + 1 )
   {
      std::cerr << "result too long\n";
      return 0;
   }

   if( ! checkcapitals(s1))
   {
      std::cerr << s1 << " contains non-capital\n";
      return 0;
   }

   if( ! checkcapitals(s2))
   {
      std::cerr << s2 << " contains non-capital\n";
      return 0;
   }

   if( ! checkcapitals(s3))
   {
      std::cerr << s3 << " contains non-capital\n";
      return 0;
   }


   ////////////////////////////
   // We first axiomatize addition in general.
   //
   //     X2 X1 X0
   //     Y2 Y1 Y0
   //   ==========
   //     Z2 Z1 Z0
   //  
   //   Ci denotes the carry from i-th to (i+1)-th column.



   for( unsigned int i = 0; i < min_l1l2; ++ i )
   {
      for( unsigned int x = 0; x < 10; ++ x )
         for( unsigned int y = 0; y < 10; ++ y )
         {
            if( i == 0 )
               writesumclauses( i, x, y, -1 );
	    else
            { 
               writesumclauses( i, x, y, 0 );
               writesumclauses( i, x, y, 1 );
            }
         }
   }
	   
   std::cout << "\n";

   for( unsigned int i = min_l1l2; i < max_l1l2; ++ i )
   {
      if( i >= l1 )
      {
         for( unsigned int y = 0; y < 10; ++ y )
         {
            writesumclauses( i, -1, y, 0 );
	    writesumclauses( i, -1, y, 1 );
         }
      }
      else
      {
         for( unsigned int x = 0; x < 10; ++ x )
         {
            writesumclauses( i, x, -1, 0 );
	    writesumclauses( i, x, -1, 1 );
         }
      }
   }

   std::cout << "\n";

#if 1
   for( unsigned int i = max_l1l2; i < l3; ++ i )
   {
      writesumclauses( i, -1, -1, 0 );
      writesumclauses( i, -1, -1, 1 );
   }
#endif

   std::cout  << "\n";

   // Last outgoing carry must be 0.

   std::cout << "[ " << "C" << ( l3 - 1 ) << "0" << " ]\n";

   std::cout << "\n\n\n";

   // We do a couple of things at the same time.
   // We produce two lists of pairs VAR VAlue.

   std::map< char, std::list< std::string > > variables; 

   for( unsigned int i = 0; i < l1; ++ i )
   {
      std::string x = "X";
      x += ( '0' + i );
      if( i + 1 < l1 )
      {
         writeatleastone( x, 0, 10 );
         writeatmostone( x, 0, 10 );
      }
      else
      {
	 // Last (= leftmost) digit should be nonzero:

         writeatleastone( x, 1, 10 );
	 writeatmostone( x, 1, 10 );
      }

      variables[ s1[ l1 - i - 1 ] ]. push_back(x);
   }

   for( unsigned int i = 0; i < l2; ++ i )
   {
      std::string y = "Y";
      y += ( '0' + i ); 
      if( i + 1 < l2 )
      {
         writeatleastone( y, 0, 10 );
         writeatmostone( y, 0, 10 );
      }
      else
      {
         // Last (= leftmost) digit should be nonzero.

         writeatleastone( y, 1, 10 );
	 writeatmostone( y, 1, 10 );
      }

      variables[ s2[ l2 - i - 1 ] ]. push_back(y);
   }

   for( unsigned int i = 0; i < l3; ++ i )
   {
      std::string z = "Z";
      z += ( '0' + i );
      if( i + 1 < l3 )
      {
         writeatleastone( z, 0, 10 );
         writeatmostone( z, 0, 10 );
      }
      else
      {
         // Last (= leftmost) digit should be nonzero.

         writeatleastone( z, 1, 10 );
	 writeatmostone( z, 1, 10 );
      }

      variables[ s3[ l3 - i  - 1 ] ]. push_back(z); 
   }

   std::cout << "\n";

#if 1 
   for( std::map< char, std::list< std::string > > :: const_iterator
           p = variables. begin( );
	   p != variables. end( );
	   ++ p )
   {
      std::cerr << ( p -> first ) << ": [ ";
      for( std::list< std::string > :: const_iterator 
              q = ( p -> second ). begin( );
	      q != ( p -> second ). end( );
	      ++ q )
      {
         if( q != ( p -> second ). begin( ))
            std::cerr << ", ";
	 std::cerr << *q;
      }
      
      std::cerr << " ]\n";
   }
#endif

   std::cout << "\n";

   // Points in the addition that contain the same letter 
   // must be equal.

   for( std::map< char, std::list< std::string > > :: const_iterator
           p = variables. begin( );
	   p != variables. end( );
	   ++ p )
   {
      for( std::list< std::string > :: const_iterator
              q1 = ( p -> second ). begin( );
	      q1 != ( p -> second ). end( );
	      ++ q1 )
      {
         for( std::list< std::string > :: const_iterator
            q2 = ( p -> second ). begin( );
	    q2 != q1;
	    ++ q2 )
         {
            writeequalityclauses( *q1,  *q2 );
         }
      }
   }

   std::cout << "\n";

   // Points that contain different letters must be different.

   for( std::map< char, std::list< std::string > > :: const_iterator
           p1 = variables. begin( );
	   p1 != variables. end( );
	   ++ p1 )
   {
      for( std::map< char, std::list< std::string > > :: const_iterator
              p2 = variables. begin( );
	      p2 != p1;
	      ++ p2 )
      {
         writedifferenceclauses( p1 -> second. front( ), 
			         p2 -> second. front( ));
      }
   }

}

