

#ifndef CLAUSESET_INCLUDED
#define CLAUSESET_INCLUDED   1


#include "clause.h"
#include "interpretation.h"


// There is a parser for clausets that can throw this error:

class syntaxerror
{
   std::string reason;
   unsigned int linenumber;

public: 
   syntaxerror( const std::string& reason, unsigned int linenumber )
      : reason( reason ), linenumber( linenumber )
   { }

   friend std::ostream& operator << ( std::ostream& , 
		                      const syntaxerror& se );

};


inline std::ostream& operator << ( std::ostream& stream,
		                   const syntaxerror& se )
{
   stream << "Syntax Error: " << se. reason << " in line ";
   stream << se. linenumber << "\n\n";
   return stream;
}



struct clauseset
{

   // A clause and its derivation.

   struct clausederiv
   {
      clause cl;
      std::list< unsigned int > deriv;
         // Input clauses have deriv. size( ) == 0.
	 // derived clauses have deriv. size( ) == 2.

      clausederiv( const clause& cl,
                   const std::list< unsigned int > & deriv )
         : cl( cl ), deriv( deriv )
      { }
   };


   std::vector< clausederiv > clauses;



   void insert( const clause& cl );
   void insert( const clause& cl, 
		unsigned int par1, unsigned int par2 );
      // Insert an input or a resolvent clause. The functions
      // are distinct, because we log the derivation.

   void remove( const clause& cl );
      // Remove all occurrences of cl.

   unsigned int size( ) const;


   typedef std::vector< clausederiv > :: const_iterator const_iterator;
   const_iterator begin( ) const { return clauses. begin( ); }
   const_iterator end( ) const { return clauses. end( ); }


   friend std::ostream& operator  << ( std::ostream& stream,
		                       const clauseset& cl );

   void read( std::istream&, unsigned int& linecount ) throw( syntaxerror );
};


#endif


